from typing import Dict, Tuple

import torch.nn as nn
from torch import Tensor
from torch import device as Device

from ._log import log_message

total_size_submodules = 0
submodules: Dict[int, Tuple[str, nn.Module]] = {}


def apply_pipeline(model: nn.Module, device: Device, method: str = "simple"):
    r"""
    Module Pipeline
    ---
    Example:
        ```python
        >>> import torch
        >>> import mbp
        >>> device = torch.device("cuda:0")
        >>> model = ... # Define your model
        >>> model = mbp.apply_pipeline(model, device)
        ```
    """
    if method == "simple":
        return apply_pipeline_simple(model, device)
    elif method == "preload":
        return apply_pipeline_preload(model, device)
    else:
        raise NotImplementedError(f"Method `{method}` is not implemented!")


def move_to_gpu(name: str, device: Device, next_mod_index: int | None = None):
    def _wrapper(module: nn.Module, *args, **kwargs):
        if next_mod_index is not None:
            if next_mod_index < total_size_submodules:
                next_mod_name, next_module = submodules[next_mod_index]
                next_module.to(device, non_blocking=True)
                log_message(
                    f">>> [Forward (pre-load)] Move `{next_mod_name}` to {next(module.parameters()).device}!",
                    color="green",
                )
            if next_mod_index - 1 == 0:
                module.to(device, non_blocking=True)
                log_message(
                    f">>> [Forward] Move `{name}` to {next(module.parameters()).device}!",
                    color="green",
                )
        else:
            module.to(device, non_blocking=True)
            log_message(
                f">>> [Forward] Move `{name}` to {next(module.parameters()).device}!",
                color="green",
            )

    return _wrapper


def move_to_cpu(name: str, next_mod_index: int):
    def _auto_onload_offload(
        name: str, module: nn.Module, device: Device, next_mod_index
    ):
        def _wrapper(*args, **kwargs):
            if next_mod_index < total_size_submodules:
                next_mod_name, next_module = submodules[next_mod_index]
                next_module.to("cpu", non_blocking=True)
                log_message(
                    f"<<< [Backward] Move `{next_mod_name}` to {next(module.parameters()).device}!",
                    color="blue",
                )
            module.to(device, non_blocking=True)
            log_message(
                f">>> [Backward] Move `{name}` to {next(module.parameters()).device}!",
                color="blue",
            )

        return _wrapper

    def _wrapper(module: nn.Module, input: Tensor, output: Tensor):
        output.grad_fn.register_prehook(
            _auto_onload_offload(name, module, output.device, next_mod_index)
        )
        module.to("cpu", non_blocking=True)
        log_message(
            f"<<< [Forward] Move `{name}` to {next(module.parameters()).device} (attach backward hooks: {hex(id(output))})!",
            color="green",
        )

    return _wrapper


def apply_pipeline_simple(model: nn.Module, device: Device):
    global submodules, total_size_submodules
    leaf_modules = []
    mod_index = 0
    for name, _ in model.state_dict().items():
        name_list = name.split(".")[:-1]
        mod: nn.Module = model
        for n in name_list:
            mod = getattr(mod, n)

        name = ".".join(name_list)
        if name not in leaf_modules:
            if not mod._parameters:
                continue
            leaf_modules.append(name)
            submodules[mod_index] = (name, mod)
            mod.register_forward_pre_hook(move_to_gpu(name, device))
            mod.register_forward_hook(move_to_cpu(name, mod_index + 1))
            mod_index += 1
    total_size_submodules = len(submodules)
    return model


def apply_pipeline_preload(model: nn.Module, device: Device):
    global submodules, total_size_submodules
    leaf_modules = []
    mod_index = 0
    for name, _ in model.state_dict().items():
        name_list = name.split(".")[:-1]
        mod: nn.Module = model
        for n in name_list:
            mod = getattr(mod, n)

        name = ".".join(name_list)
        if name not in leaf_modules:
            if not mod._parameters:
                continue
            leaf_modules.append(name)
            submodules[mod_index] = (name, mod)
            mod.register_forward_pre_hook(move_to_gpu(name, device, mod_index + 1))
            mod.register_forward_hook(move_to_cpu(name, mod_index + 1))
            mod_index += 1
    total_size_submodules = len(submodules)
    return model



import argparse
import time
from contextlib import contextmanager

import torch
import torch.nn as nn
import torch.optim as optim
from torchvision.models import ResNet50_Weights, resnet50

import mbp


@contextmanager
def measure_time(method_name: str, runtimes: list):
    stime = time.perf_counter()
    yield
    etime = time.perf_counter()
    runtime = etime - stime
    runtimes.append(runtime)
    print(f"[{method_name}] Time: {runtime:.1f} (sec)")


def get_arguments():
    parser = argparse.ArgumentParser()
    # Dataset
    parser.add_argument("--seed", type=int, default=42)
    parser.add_argument("--image_size", type=int, default=224)
    parser.add_argument("-b", "--batch_size", type=int, default=16)
    parser.add_argument("--num_class", type=int, default=1000)
    parser.add_argument("--num_workers", type=int, default=0)

    parser.add_argument("--device", type=int, default=0)

    parser.add_argument("--epochs", type=int, default=10)
    parser.add_argument("--lr", type=float, default=0.1)
    parser.add_argument("--weight_decay", type=float, default=1e-4)
    parser.add_argument("--momentum", type=float, default=0.9)

    parser.add_argument("--method", type=str, default="default", help="default, mbp")
    parser.add_argument("-u", "--micro_batch", type=int, default=8)
    parser.add_argument("--dp", action="store_true")

    return parser.parse_args()


if __name__ == "__main__":
    args = get_arguments()

    torch.manual_seed(args.seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

    dev = torch.device(f"cuda:{args.device}")
    model = resnet50(weights=ResNet50_Weights.IMAGENET1K_V1)
    model = (
        nn.DataParallel(model) if args.dp else model
    )  # TODO: Check if this is correct
    model.to(dev)

    criterion = nn.CrossEntropyLoss()
    optimizer = optim.SGD(
        model.parameters(),
        lr=args.lr,
        momentum=args.momentum,
        weight_decay=args.weight_decay,
    )

    @mbp.apply(["input", "target"], args.micro_batch)
    def train_fn(model, criterion, input, target):
        output = model(input)
        loss = criterion(output, target)
        return loss, output, output.shape[0]

    avg_runtime = []
    if args.method == "mbp":
        with measure_time(args.method, avg_runtime):
            for epoch in range(args.epochs):
                input, target = torch.randn(
                    args.batch_size, 3, args.image_size, args.image_size
                ), torch.randint(0, args.num_class, (args.batch_size,))
                optimizer.zero_grad()
                loss, output, num = train_fn(
                    model, criterion, input=input, target=target
                )
                print(loss, output.shape, num)
                optimizer.step()
    else:
        with measure_time(args.method, avg_runtime):
            for epoch in range(args.epochs):
                input, target = torch.randn(
                    args.batch_size, 3, args.image_size, args.image_size
                ).to(dev), torch.randint(0, args.num_class, (args.batch_size,)).to(dev)
                optimizer.zero_grad()
                output = model(input)
                loss = criterion(output, target)
                loss.backward()
                print(loss.detach().item())
                optimizer.step()



import functools
import math
import warnings
from typing import Any, Callable, Dict, List, Tuple

import torch
from torch import Tensor
from torch import device as Device

from ._log import runtime_
from ._pipeline import apply_pipeline


def apply(
    batch_names: List[str] | Tuple[str, ...],
    ub_size: int = 1,
    device: str | int | Device = "cpu",
    device_ids: List[int] | Tuple[int, ...] | None = None,
) -> Callable:
    r"""
    MBP Decorator
    ---
    The `apply()` decorator helps in training models with large batches by splitting them into smaller micro-batches.
    This decorator calculates the loss and performs backpropagation to accumulate gradients without updating model parameters.

    Usage:
    1. Specify the tensor names to be split and the micro-batch size.
    2. Ensure the decorated function computes and returns the loss without calling `.backward()`.
    3. Explicitly define the corresponding tensors in `key=value` format (e.g., `x=image`) to apply MBP.
    If not, MBP will not be applied to the tensors.

    Example:
        ```python
        >>> import mbp
        >>> @mbp.apply(["x", "label"], ub_size=16) # 1
        >>> def train_fn(model, criterion, x, label, *args, **kwargs):
        ...     o = model(x)
        ...     loss = criterion(o, label)
        ...     return loss # 2
        >>> for image, label in dataloader:
        ...     optimizer.zero_grad()
        ...     loss = train_fn(model, criterion, x=image, label=label) # 3
        ...     optimizer.step()
        ```

    Args:
        batch_names (List[str] | Tuple[str]): List of batch names to be split.
        ub_size (int): Micro-batch size. Defaults to 1.
        device (str | int | Device, optional): Device to which the micro-batches will be sent. Defaults to "cpu".
        device_ids (List[int] | Tuple[int]): List of GPU IDs for multi-GPU training. Defualts to None.

    Returns:
        loss (float): The loss value.
    """
    assert isinstance(
        batch_names, (list, tuple)
    ), "`batch_names` must be a list or tuple."
    assert isinstance(
        ub_size, int
    ), "`ub_size` must be an integer or string. (Default: 1)"
    assert isinstance(
        device, (str, int, Device)
    ), "`device` must be a string, integer, or torch.device. (Default: 'cpu')"
    if isinstance(device, str):
        dev_ = Device(device)
    elif isinstance(device, int):
        dev_ = Device(f"cuda:{device}")
    else:
        dev_ = device
    if device_ids is not None:
        assert isinstance(
            device_ids, (list, tuple)
        ), "`device_ids` must be a list or tuple. (Default: 'None')"
        device_ids_ = len(device_ids)
    else:
        device_ids_ = 1

    def decorate(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not hasattr(wrapper, "dev"):
                wrapper.dev, wrapper.num_device = dev_, device_ids_
            loss = 0
            batch_chunker = BatchChunker(
                batch_names, kwargs, ub_size, wrapper.dev, wrapper.num_device
            )
            if batch_chunker.is_valid:
                ub_loss: Tensor
                others: List[Any] = []  # Store/Stack others
                for ubatch in batch_chunker:
                    kwargs.update(ubatch)
                    out = _forward_ub(func, args, kwargs)
                    ub_loss, others = _seperate_uloss_and_others(out, others)
                    loss += _backward_ub(ub_loss, batch_chunker.chunk_size)
                if len(others) > 0:
                    return loss, *_gather_others(others)
                else:
                    return loss
            else:
                warnings.warn(
                    "No tensors to split. please check the format of arguments in the function (key=value).",
                )
                return func(*args, **kwargs)

        return wrapper

    return decorate


@runtime_
def _forward_ub(func: Callable, args: List[Any], kwargs: Dict[str, Any]) -> Any:
    return func(*args, **kwargs)


@runtime_
def _backward_ub(ub_loss: Tensor, chunk_size: int) -> Tensor:
    ub_loss /= chunk_size  # Apply Loss Normalization.
    ub_loss.backward()
    return ub_loss


def _seperate_uloss_and_others(
    out: Any, others_list: List[List[Any]]
) -> Tuple[Tensor, List[List[Any]]]:
    # Separate the loss tensor and others from the output of the function.
    if isinstance(out, tuple):
        ub_loss, *others = out
        assert isinstance(
            ub_loss, Tensor
        ), "The first return value must be a loss tensor."
        assert ub_loss.requires_grad, "The loss tensor must require gradients."
        assert ub_loss.dim() == 0, "The loss tensor must be a scalar."
        others_list = _store_others(others, out=others_list)
        return ub_loss, others_list
    else:
        ub_loss = out
        assert isinstance(
            ub_loss, Tensor
        ), "The first return value must be a loss tensor."
        assert ub_loss.requires_grad, "The loss tensor must require gradients."
        assert ub_loss.dim() == 0, "The loss tensor must be a scalar."
        return ub_loss, others_list


def _store_others(others: Tuple[Any], out: List[List[Any]]) -> List[Any]:
    # Store others in a list.
    if len(out) == 0:
        out = [[o] for o in others]
    else:
        for i, o in enumerate(others):
            out[i].append(o)
    return out


def _gather_others(others: List[Any]) -> List[Any]:
    # Gather others from the list.
    for i, o in enumerate(others):
        if len(o) == 1:
            others[i] = o[0]
        else:
            if isinstance(o[0], Tensor):
                others[i] = torch.cat(o)  # Only for tensors to concatenate
            else:
                others[i] = o
    return others


class BatchChunker:
    def __init__(
        self,
        batch_names: List[str] | Tuple[str, ...],
        kwargs: Dict[str, Tensor],
        ub_size: int,
        device: Device,
        num_device: int,
    ):
        for n in batch_names:
            assert n in kwargs, f"Missing tensor: {n}"
        mb_size = kwargs[batch_names[0]].shape[0]
        chunk_size = math.ceil(mb_size / ub_size) if mb_size > ub_size else 1
        chunk_size = (
            math.ceil(chunk_size / num_device) if num_device > 1 else chunk_size
        )

        self._stop_index = chunk_size
        self._curr_index = 0
        self.device: Device = device
        self.batch_names = batch_names
        self.micro_batches: List[Dict[str, Tensor]] = []
        for s in range(chunk_size):
            sidx = s * ub_size
            eidx = min(sidx + ub_size, mb_size)
            self.micro_batches.append({n: kwargs[n][sidx:eidx] for n in batch_names})
        for n in batch_names:
            del kwargs[n]

    def __len__(self):
        return self._stop_index

    def __iter__(self):
        return self

    def __next__(self):
        if self._curr_index < self._stop_index:
            self._curr_index += 1
            return self._load_ub(self._curr_index - 1)
        else:
            self._curr_index = 0
            raise StopIteration

    @runtime_
    def _load_ub(self, _curr_index: int) -> Dict[str, Tensor]:
        ubatch = self.micro_batches[_curr_index]
        for n in ubatch:
            ubatch[n] = ubatch[n].to(self.device, non_blocking=True)
        return ubatch

    @property
    def chunk_size(self):
        return self._stop_index

    @property
    def is_valid(self):
        return self.chunk_size > 1


__all__ = [apply, apply_pipeline]
__version__ = "0.2.5"



import functools
import logging
from contextlib import contextmanager
from time import perf_counter_ns
from typing import Callable

logging.basicConfig(
    level=logging.INFO,
    # filename="panda.log",
    format="%(asctime)s [%(levelname)s] %(message)s",
)


class TextColor:
    """
    Text color class
    """

    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    RESET = "\033[0m"

    @staticmethod
    def colorize(text: str, color: str | None = None) -> str:
        """
        Colorize text
        :param text: text to colorize
        :param color: color
        :return: colorized text
        """
        color = color or TextColor.RESET
        return f"{color}{text}{TextColor.RESET}"


def log_message(txt: str, color: str = TextColor.RESET):
    r"""
    Log message
    :param txt: text to log
    :param color: color

    """
    _color = TextColor.RESET
    if _color in dir(TextColor):
        _color = getattr(TextColor, color.upper())
    logging.debug(TextColor.colorize(txt, _color))


@contextmanager
def runtime(txt: str):
    start = perf_counter_ns()
    yield
    end = perf_counter_ns()
    logging.debug(
        TextColor.colorize(
            f"<{txt}> Runtime: {(end-start) * 1e-6:.1f} ms", TextColor.MAGENTA
        )
    )


def runtime_(func: Callable):
    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        with runtime(func.__name__):
            return func(*args, **kwargs)

    return _wrapper



import sys

from setuptools import find_packages, setup

REQUIRED_PYTHON_VERSION = (3, 10)


def check_python_version():
    if sys.version_info < REQUIRED_PYTHON_VERSION:
        print(
            "Error: This package requires Python {} or higher.".format(
                ".".join(map(str, REQUIRED_PYTHON_VERSION))
            )
        )
        sys.exit(1)


def check_pytorch_cuda():
    try:
        import torch

        if not torch.cuda.is_available():
            print("Error: PyTorch is installed but CUDA is not available.")
            sys.exit(1)
    except ImportError:
        print("Error: PyTorch is not installed. Install it manually using:")
        print("pip install torch torchvision torchaudio")
        sys.exit(1)


check_python_version()
check_pytorch_cuda()


setup(
    name="mbp-pytorch",
    version="0.2.5",
    packages=find_packages(),
    author="XinYu Piao",
    author_email="xypiao97@korea.ac.kr",
    description="Enabling large batch size training for DNN models beyond the memory limit",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/HPIC/MBP.git",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
    ],
    python_requires=">=3.10",
    include_package_data=True,
)
